#ifndef MAPPING_H
#define MAPPING_H

/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: {mittals, pdmcdan}@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : mapping.h
//
//   Created : 14-Aug-00
//   By : Sharad Mittal
//
//   Description : This is an internal declaration file that declares a 
//                 a class for storing all the socket handles.
//
//   *** Change History ***
//
//   14-Nov-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////


// Constant Definitions used for error handling
#define INVALID_SOCKET 1
#define UNKNOWN_ERROR 2
#define ERR_GROUP 3
#define GRP_JOIN 4
#define MAXDATA 1024
#define SEND_ERR 5
#define ERR_READ 6
#define SOCK_ERR 7
#define ERR_NOGROUP 8

#define PE_DEFAULT_ALEVEL DEFAULT_ALEVEL
#define SEND_PERIOD 5*1000

// Class SockObject Definition : stores information related to one socket
class SockObject {
 public:
     AGroup *grp;  //Group Object 
     int type;     // Type of the Socket: Datagram, Stream, ..
     int port;     // Port  to which the socket is bound (if bound)
     int recv_sockpair[2]; /* This sockpair is used for inter thread 
                              communication on receiving a packet */
     int send_sockpair[2]; /* This sockpair is used for inter thread
                              communication when a packet is to be sent */
     char *pol;    // this is the policy filename
     
     SockObject() {}; // empty constructor
};

// Class Listmap Definition : this class maintains all the sockets created
class Listmap {
 public:
     Listmap() { //Constructor
               mp = NULL; 
               maxfd = 0;
          }
     // This method inserts a new group object into the list of sockhandles 
     // and returns a "socket" handle for it. "type" is the socket type.
     int Insert(AGroup *group,int type, const char *filename)
          {
               int fd=++maxfd;
               int filelen = strlen(filename);

               List *temp = new List();
               SockObject *sk = new SockObject();
               sk->grp = group;
               sk->type = type;
               sk->port = -1;
               
               if ((filelen < 1)&& (filelen > 100)) 
                    sk->pol = NULL;
               else {
                    // do a malloc
                    sk->pol = (char *)malloc(filelen+1);
                    memset(sk->pol,'\0',filelen);
                    strcpy(sk->pol,(char *)filename);
               }
               //temp = new struct list;
               if (temp == NULL) return(-1);
               temp->fd = fd;
               temp->sock = sk;
               temp->next = mp;
               mp = temp;
               return(fd);
          }

     // maps a socket handle to the corresponding SockObject Entry
     SockObject *mapfd(int fd)
          {
               List *temp = mp;
               while (temp!=NULL) {
                    if (temp->fd == fd)
                         return (temp->sock);
                    temp=temp->next;
               }
               return (NULL);
          }

     // to change the group corresponding to a particular socket
     int changegp(int fd, AGroup *group)
          {
               List *temp = mp;
               while (temp!=NULL) {
                    if (temp->fd == fd) { 
                         temp->sock->grp  = group;
                         return(0);
                    }
                    temp=temp->next;
               }
               return (-1);
          }

     // deletes an entry indexed by handle "dfd" from the sockhandles list
     int del(int dfd)
          {
               List *curr = mp,*prev = mp;
               while ((curr!=NULL)&&(curr->fd!=dfd)) {
                    prev=curr;
                    curr = curr->next;
               }
               if (curr == NULL) return(-1);
               if (curr == prev) {
                    mp = mp->next;
                    prev = NULL;
                    delete curr;
               }
               else {
                    prev->next = curr->next;
                    delete curr;
               }
               return(0);
          }


     private :
          class List {
          public:
          int fd;
          SockObject *sock;
          List *next;

          List(){}; //void constructor
     };
     List *mp;
     int maxfd;
};

#endif








