/******************************************************************
 *
 * Software Systems Research Laboratory
 * Department of Electrical Engineering and Computer Science
 * University of Michigan at Ann Arbor
 * 1301 Beal Avenue, Ann Arbor, MI 48105
 * 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: {mittals, pdmcdan}@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : socket.cpp
//
//   Created : 14-Nov-00
//   By : Sharad Mittal
//
//   Description : (See socket_s.h for function declarations) This is the main src
//                 file for Socket_s library.
//
//   *** Change History ***
//
//   14-Nov-00 Created File : DPM
//   Last Modified: 19-Jul-01
//
//////////////////////////////////////////////////////////////////////////////////


#define SHARAD_SOURCE

// define the Antigone broadcast layer operation mode here. Set it to
// ATransport::AT_SYMMETRIC (Symmetric), ATransport::AT_ASYMMETRIC
// (Asymmetric) or ATransport::AT_PTTOPT (Point-to-point).
#define MODE ATransport::AT_SYMMETRIC

// define the server address for the IP address of the machine running
// the Antigone server. For the Symmetric mode of operation, you can
// leave it as blank. For example,
// #define SADDR "192.168.27.3"
// you can also define the environment variable SADDR
#define SADDR_DEF ""

// define the server port here, although it will not be used in
// Symmetric mode and Asymmetric mode. It can also be defined as an
// environment variable SPORT
#define SPORT_DEF 5001

#define DEFAULT_POLICY "loc.apd"
#define MAXMSG 1024

// Include Files
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>

// Antigone Include Files
#include <AGroup.h>
#include <AtkThread.h>
#include <AtkTimer.h>
#include <AtkThreadLock.h>
#include <Environment.h>

//mapping.h is a local include file
#include <mapping.h>
#include <socket_s.h>

// Global variables for this file
AtkThreadLock *lk; 
//char pol[50]="loc.apd";
int errorno;
  
class Listmap sockhandles;
AuditLog *alog;

//////////////////////////////////////////////////////////////////////////////////
//
// Function     : socket_s
// Description  : socket constructor
//
// Inputs       : family - domain of the socket (ex - AF_UNIX, ..)
//                type - type of the desired socket (STREAM, DATAGRAM, ..)
//                protocol - protocol to be used by the socket
// Outputs      : socket handle for future references

int socket_s(int family, int type, int protocol)
{
     char *alogs;
     if (family!=AF_INET) return(-1);

     // currently Socket_s library only supports multicast 
     if (type!=SOCK_DGRAM) {
          errorno = SOCK_ERR;
          return(-1);
     }

     // valid protocol checking
     if ((protocol!=IPPROTO_IP)&&(protocol!=IPPROTO_TCP)&&(protocol!=IPPROTO_UDP))
          return(-1);

     //AGroup *group = NULL;
     int fd = sockhandles.Insert(NULL,type,DEFAULT_POLICY);

     //  log creation for logging all the debug information
     alogs = getenv("NAME");
     alog = new AuditLog( alogs, FALSE, TRUE , TRUE);
     alog->setAuditLevel( PE_DEFAULT_ALEVEL );

     //saddr = (unsigned long)inet_addr("192.168.27.1");
     return(fd);
}

//////////////////////////////////////////////////////////////////////////////////
//
// Procedure    : comm_s
// Description  : This is the thread procedure spawned by the main process. It 
//                takes care of all the communication with the Antigone server
//                and other members of the group.
//
// Inputs       : sockfd - it is the pointer to the socket handle

// Outputs      : None

void *comm_s(int *sockfd)
{
     String snder;
     int n;
     char dat[10];
     //char data1[MAXDATA];
     //char type[4];
     int sfd = *sockfd,i=0;
     unsigned long idx = 1, rindx;
     char data[MAXMSG];
     unsigned long speriod = 10;

     LOGMSG0( alog, PE_DEFAULT_ALEVEL, "comm thread started");
     printf("comm thread started\n");
     // getting the corresponding socket info 
     SockObject *sk;
     if ((sk = sockhandles.mapfd(sfd)) == NULL) {
          errorno = ERR_GROUP;
          LOGMSG0(alog, PE_DEFAULT_ALEVEL,"comm_s: sockhandle error");
          return NULL;
     }
  
     AtkTimer read_timer(speriod);
     while ( sk->grp->State() == AGroup::AG_INOPERATION ){
          Buffer *buf = NULL;
          int inloop=0;

          //pthread_mutex_lock(&ft);
          LOGMSG0( alog, PE_DEFAULT_ALEVEL, "before reading a message");

          read_timer.reset();
          if ( (sk->grp->readMessage( &buf , &read_timer) == TRUE) 
               && (buf != NULL) )
          {
               LOGMSG0( alog, PE_DEFAULT_ALEVEL, "just read a message");  
               //pthread_mutex_unlock(&ft);
               // Read message, log it
               LOGMSG0( alog, PE_DEFAULT_ALEVEL, "decoding the message");

               int dlen = buf->Length();
               LOGMSG1(alog, PE_DEFAULT_ALEVEL,"recvd pkt length = %d",dlen);
               fflush(stdout);
               LOGMSG1(alog, PE_DEFAULT_ALEVEL,"message length = %d",dlen);

               if (dlen>0) buf->getFixedLength((unsigned char *)&dat,dlen);
               LOGMSG1(alog, PE_DEFAULT_ALEVEL,"message = %s",dat);

               // sending the received data to the local socket
               if ((dlen>0)&&(write(sk->recv_sockpair[1],dat,dlen) != dlen))
                    printf("socket error");
               delete buf;
          }
          // sending 
          LOGMSG0(alog, PE_DEFAULT_ALEVEL, "about to send a message");
          memset((char *)&data,'\0',MAXMSG);
          n = read(sk->send_sockpair[0], &data, MAXMSG);
          if (n > 0) {
               Buffer *buf = new Buffer();
               buf->putFixedLength((unsigned char *) data,n);
               sk->grp->sendMessage( buf );
               LOGMSG0(alog, PE_DEFAULT_ALEVEL,"comm_s: Sent a APP packet");
               //printf("comm_s: Sent a APP packet\n");
          }

          //printf("comm_s: loop completed\n");
          LOGMSG0(alog, PE_DEFAULT_ALEVEL,"comm_s: loop completed");

          //pthread_mutex_unlock(&ft);
          usleep(1);
     }
     LOGMSG0(alog, PE_DEFAULT_ALEVEL,"group not connected now");
     //printf("group not connected now\n");
}

//////////////////////////////////////////////////////////////////////////////////
//
// Procedure    : bind_s
// Description  : bind a socket to a port and an interface
//
// Inputs       : sockfd - socket handle
//                my_addr - address data structure with port and interface address
//                addrlen - length of the address data structure
// Outputs      : procedure status

int bind_s(int  sockfd, struct sockaddr_in *my_addr, int addrlen)
{
     SockObject *sk;
     if ((sk = sockhandles.mapfd(sockfd)) == NULL)
          return(-1);

     if (addrlen != sizeof(struct sockaddr_in)) 
          return(-1);
     sk->port = ntohs(my_addr->sin_port);
}

//////////////////////////////////////////////////////////////////////////////////
//
// Function     : join_mult
// Description  : joins a socket to a multicast group (using Antigone). Internal
//                function.
//
// Inputs       : sockfd - socket handle
//                mult_addr - multicast address data structure for multicast group
//
// Outputs      : procedure status

int join_mult(int  sockfd, struct ip_mreq *mult_addr)
{
     SockObject *sk;
     char *saddr = getenv("SADDR");
     char *port_str = getenv("SPORT");
     int sport;
     ATransport *trans;

     if (sockfd < 0) {
          errorno = INVALID_SOCKET;
          return (-1);
     }

     if ((sk = sockhandles.mapfd(sockfd)) == NULL)
          return(-1);

     String *lpol=new String(sk->pol,-1);
     LOGMSG1(alog,PE_DEFAULT_ALEVEL,"The policy file is %s",sk->pol);
     
     ATransport::AtMode mode = MODE;
     char *modenv = getenv("MODE");

     if (strcmp(modenv,"ASYM") == 0) {
          mode = ATransport::AT_ASYMMETRIC;
          LOGMSG0(alog, PE_DEFAULT_ALEVEL, "MODE is ASYMMETRIC");
     }
     if (strcmp(modenv,"PTOP") == 0) {
          mode = ATransport::AT_PTTOPT;
          LOGMSG0(alog, PE_DEFAULT_ALEVEL, "MODE is PTTOPT");
     }
     if (mode == MODE) 
          LOGMSG0(alog, PE_DEFAULT_ALEVEL, "MODE is SYMMETRIC");

     IPAddress *ipsaddr = NULL;

     if (mode != ATransport::AT_SYMMETRIC)
          ipsaddr = IPAddress::IPAddressFactory(saddr , 5001);

     int portno = sk->port;
     if (portno < 1)
          portno = 5000; // default port allocation

     if (port_str != NULL) 
          sport = atoi(port_str);
     else // no environment definition
          sport = SPORT_DEF;
     
     if (saddr == NULL) // no environment definition
          saddr = SADDR_DEF;

     // create a symmetric transport object

     IPAddress *lipaddr;
     ATransport::SetLoopback( TRUE );
     
     LOGMSG0(alog, PE_DEFAULT_ALEVEL,"setting up the IPAddress structures");

     if (mode == ATransport::AT_PTTOPT)
          lipaddr = IPAddress::IPAddressFactory( saddr, sport);
     else
          lipaddr = IPAddress::IPAddressFactory( mult_addr->imr_multiaddr.s_addr , portno );

     if ( lipaddr == NULL )
     {
          // Error out
          LOGMSG0(alog,PE_DEFAULT_ALEVEL,"join_mult: IPAddress Error"); 
          errorno = UNKNOWN_ERROR;
          return( -1 );
     }
  
     if ( mode == ATransport::AT_PTTOPT )
          // unicast address is used as (virtual) mcast address on P2P
          trans = new ATransport( (IPAddress *)NULL, lipaddr, mode );
     else
          trans = new ATransport( lipaddr, ipsaddr, mode );
  
     AGroup *grp = new AGroup(trans, *lpol, NULL, NULL, alog);
     trans->Log() = alog;
     //grp->LocalPolicy()->logPolicy( PE_DEFAULT_ALEVEL, alog );

     // setting up the socket handle mapping
     if (sockhandles.changegp(sockfd,grp) < 0) {
          errorno = ERR_GROUP;
          return(-1);
     }

     LOGMSG0(alog, PE_DEFAULT_ALEVEL, "connecting to the group");
     if ( grp->Connect() != TRUE )
     {
          // Error out
          LOGMSG0( alog, FATAL_ERROR, "Error connecting/joining group." );
          errorno = GRP_JOIN;
          return(-1);
     }
     //SockObject *sk;
     if ((sk = sockhandles.mapfd(sockfd)) == NULL) {
          errorno = ERR_GROUP;
          return (-1);
     }
  
     //creating a socketpair for communication with thread "comm_s"
     if (socketpair(AF_UNIX, SOCK_DGRAM, 0, sk->recv_sockpair) < 0) {
          errorno = ERR_GROUP;
          return(-1);
     }

     if (socketpair(AF_UNIX, SOCK_DGRAM, 0, sk->send_sockpair) < 0) {
          errorno = ERR_GROUP;
          return(-1);
     }
  
     // Setting reading socket of the pair to be non-blocking
     fcntl(sk->recv_sockpair[0],F_SETFL, O_NONBLOCK);
     fcntl(sk->send_sockpair[0],F_SETFL, O_NONBLOCK);

     static int sfd=sockfd;
     int cpd;
     if ((cpd = fork()) < 0) {
          LOGMSG0(alog, PE_DEFAULT_ALEVEL, "(join_mult): fork failed");
          printf("fork failed\n");
          exit(-1);
     }
     if (cpd == 0) {
          comm_s(&sfd);
          return(1);
     }
     //AtkThread *th = new AtkThread((AtkThreadFuncPtr) &comm_s, &sfd); 
     unsigned int in=1; 

     lk = new AtkThreadLock();
     LOGMSG0(alog, PE_DEFAULT_ALEVEL," starting the comm_s thread");
     //comm(&sfd);
     //th->Launch();
}

//////////////////////////////////////////////////////////////////////////////////
//
// Function     : setsockopt
// Description  : used to set socket options (joining a multicast group, TTL, ..
//
// Inputs       : sockfd - socket handle
//                level - socket options level 
//                optname - option name
//                optval - corresponding value for the option
//                optlen - length of the option value
//                (refer setsockopt manual page for more information on parameters.
//
// Outputs      : procedure status


int setsockopt_s(int sockfd, int level, int optname,  const  void
                 *optval, socklen_t optlen)
{
     struct ip_mreq *madd = (struct ip_mreq *)optval;
     switch (optname) {
     case IP_ADD_MEMBERSHIP: // add the socket to a multicast group
          if (optlen != sizeof(struct ip_mreq))
               return(-1);
          return(join_mult(sockfd,madd));
          break;

     default: 
          LOGMSG1(alog, PE_DEFAULT_ALEVEL,"setsockopt_s: Option not supported:%d",optname);
          break;
     }
}

unsigned long idx = 0;
//////////////////////////////////////////////////////////////////////////////////
//
// Function     : sendto_s
// Description  : used to send data to particular destination. 
//
// Inputs       : sockfd - socket handle
//                msg - message to be transmitted
//                len - message length
//                flags - used to send sendto options
//                to - address to which data should be sent
//                tolen - length of the "to" data structure
//
// Outputs      : procedure status

int  sendto_s(int sockfd, const void *msg, size_t len, int flags,
              const struct sockaddr *to, socklen_t tolen)
{
     //String type = "DATA";
     /* don't need to explicitly create a String
     static unsigned char dat[MAXDATA];
     bzero(dat,MAXDATA);
     memcpy(dat,msg,len);
     String *data = new String(dat,len);*/

     Buffer *buf = new Buffer();
     SockObject *sk;
     if ((sk = sockhandles.mapfd(sockfd)) == NULL) {
          errorno = ERR_GROUP;
          return (-1);
     }
     //pthread_mutex_lock(&ft);
     AGroup *grp = sk->grp;
     
     lk->Lock();
     if (grp->State() != AGroup::AG_INOPERATION ){
          lk->Unlock();
          printf("group not connected: server down");
          return(-1);
     }
     lk->Unlock();
     /*
    char hostnm[10];
    bzero(hostnm,10);
    gethostname(hostnm,10);
    struct hostent *hs = gethostbyname(hostnm); 
    struct sockaddr_in *haddr = (struct sockaddr_in *)hs->h_addr_list[0];
    char *p=inet_ntoa((struct in_addr)haddr->sin_addr);
  */

     printf("data length before the message = %d\n",buf->Length());
     buf->putFixedLength(/*dat*/ (unsigned char *) msg,len);
     printf("data length = %d\n",buf->Length());
  
     // new changes 06/10/01

     lk->Lock();
     grp->sendMessage( buf );
     LOGMSG0(alog, PE_DEFAULT_ALEVEL, "sendto_s: sent a message");

     //delete data;
     //delete sk;

     lk->Unlock();
     //pthread_mutex_unlock(&ft);
     //delete buf;
     return(1);
}

//////////////////////////////////////////////////////////////////////////////////
//
// Function     : send_s
// Description  : used to send data to a connected host or group. 
//
// Inputs       : sockfd - socket handle
//                msg - message to be transmitted
//                len - message length
//                flags - used to send send options
//
// Outputs      : procedure status


int send_s(int s, const void *msg, size_t len, int flags)
{
     SockObject *sk;
     if ((sk = sockhandles.mapfd(s)) == NULL) {
          errorno = ERR_GROUP;
          printf("send_s: couldn't find the group");
          LOGMSG0(alog, PE_DEFAULT_ALEVEL,"send_s: couldn't find the group");
          return (-1);
     }
     
     write(sk->send_sockpair[1],msg,len);
} 

//////////////////////////////////////////////////////////////////////////////////
//
// Function     : send_s
// Description  : used to send data to a connected host or group. 
//
// Inputs       : sockfd - socket handle
//                msg - message to be transmitted
//                len - message length
//                flags - used to send send options
//
// Outputs      : procedure status

/*
  int send_s(int s, const void *msg, size_t len, int flags)
  {
  Buffer *buf = new Buffer();
  SockObject *sk;
  if ((sk = sockhandles.mapfd(s)) == NULL) {
  errorno = ERR_GROUP;
  printf("send_s: couldn't find the group");
  LOGMSG0(alog, PE_DEFAULT_ALEVEL,"send_s: couldn't find the group");
  return (-1);
  }
  //pthread_mutex_lock(&ft);
  //lk->Lock();
  AGroup *grp = sk->grp;
  
  lk->Lock();
  if (grp->State() != AGroup::AG_INOPERATION ){
  lk->Unlock();
  printf("group not connected: server down");
  LOGMSG0(alog, PE_DEFAULT_ALEVEL, "send_s: group not connected");
  return(-1);
  }
  lk->Unlock();
  printf("data length before the message = %d\n",buf->Length());
  buf->putFixedLength(dat (unsigned char *)msg,len);
  printf("data length = %d\n",buf->Length());
  
// new changes 06/10/01
lk->Lock();
grp->sendMessage( buf );

lk->Unlock();
LOGMSG0(alog,PE_DEFAULT_ALEVEL, "send_s: just sent a message");

return(1);
}
*/
//////////////////////////////////////////////////////////////////////////////////
//
// Function     : recvfrom_s
// Description  : used to send data to particular destination. 
//
// Inputs       : sockfd - socket handle
//                buf - previously allocated space for received message
//                len - maximum message length
//                flags - used to set recvfrom_s options
//                from - address from which data is received
//                tolen - length of the from data structure after it is filled
//
// Outputs      : procedure status

int  recvfrom_s(int  sockfd,  void  *buf,  size_t len, int flags,
                struct sockaddr *from, socklen_t *fromlen)
{
     SockObject *sk;
     int n;
     if ((sk = sockhandles.mapfd(sockfd)) == NULL) {
          errorno = ERR_GROUP;
          return (-1);
     }

     if ((n = read(sk->recv_sockpair[0], buf, len)) < 0) {
          //LOGMSG1(alog, PE_DEFAULT_ALEVEL, 
          //        "recvfrom_s: Error: received %d amount of data",n);
          errorno = ERR_READ;
          return(-1);
     }
     if (n > 0) {
       LOGMSG1(alog, PE_DEFAULT_ALEVEL, 
               "recvfrom_s: received %d amount of data",n);
       printf("received data: %d\n",n);
     }

     /* these lines provide the joiner identity and have been commented
        for now; they require support in the antigone code. 

        printf("len recvd in recvfrom_s = %d\n",n);
        struct sockaddr_in *frm = (struct sockaddr_in *)from;
        unsigned long frmaddress;
        unsigned short port;
        char *ptr = (char *)buf;
        ptr += n-6; // getting the address & the port
        memcpy((void *)&frmaddress, (void *)ptr, sizeof(unsigned long));
        ptr+=sizeof(unsigned long);
        memcpy((void *)&port, (void *)ptr, sizeof(unsigned short));
        struct in_addr in;
        in.s_addr = frmaddress;
        char *ip = inet_ntoa(in);
        printf("sendr2 ip = %s, port = %d\n",ip,port);
     
        if (frm != NULL) {
        frm->sin_family       = AF_INET;
        frm->sin_addr.s_addr  = frmaddress;
        frm->sin_port         = htons(port);
        }
     */
     return(n);
}

//////////////////////////////////////////////////////////////////////////////////
//
// Function     : close_s
// Description  : used to close a socket.
//
// Inputs       : sockfd - socket handle
//
// Outputs      : procedure status


int close_s(int sockfd)
{
     SockObject *sk;
     int n;
     printf("close_s: closing the Antigone Group\n");
     if ((sk = sockhandles.mapfd(sockfd)) == NULL) {
          errorno = ERR_GROUP;
          return (-1);
     }

     ATransport *atrans = sk->grp->Transport();
     delete atrans;
     delete sk->grp;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Function     : setpolicy_s
// Description  : sets the default local policy of the Antigone group member
//
// Inputs       : sockfd - socket handle
//                filename - name of the policy file to be used
//
// Outputs      : procedure status


int setpolicy_s(int sockfd, const char *filename)
{
     SockObject *sk;
     if ((sk = sockhandles.mapfd(sockfd)) == NULL) {
          errorno = ERR_NOGROUP;
          return (-1);
     }
     if (sk->pol != NULL)
          delete sk->pol;
     sk->pol = new char(strlen(filename));
     strncpy((char *)sk->pol, filename, strlen(filename));
}     


