#ifndef _SOCKET_S_H
#define _SOCKET_S_H

/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: {mittals, pdmcdan}@eecs.umich.edu
*******************************************************************/

/**************************************************************************
 **
 **   File : socket_s.h
 ** 
 **   Created : 14-Nov-00 
 **   By : Sharad Mittal 
 ** 
 **   Description : This is the main header file for the Socket_s library.  
 **                 Socket_s is a user level socket library which uses the  
 **                 Antigone 
 **                 framework as its underlying layer. All the defined  
 **                 functions 
 **                 have the same semantics as the corresponding functions in  
 **                 the kernel socket library.   
 ** 
 **   *** Change History *** 
 ** 
 **   14-Nov-00 Created File : DPM 
 ** 
 *************************************************************************/ 

/*** Include Files ***/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <stdlib.h>


/* socket, bind, .. command declarations
   extern int errorno;*/

#define SEND_PERIOD 5*1000

#ifdef SHARAD_SOURCE
extern "C" {
#endif

     extern int errorno;

     /* socket creation call*/
     int socket_s(int family, int type, int protocol);
     /* binds a created "socket" to a port*/
     int bind_s(int  sockfd, struct sockaddr_in *my_addr, int addrlen); 


     /* sending messages on a socket*/
     int write_s(int sockfd, const void *buf, size_t count);  

     int send_s(int sockfd, const void *msg, size_t len, int flags);

     int sendto_s(int sockfd, const void *msg, size_t len, int flags,
                   const struct sockaddr *to, socklen_t tolen);

     int sendmsg_s(int sockfd, const struct msghdr *msg, int flags);
  


     /*receiving messages on a socket*/
     int read_s(int sockfd, void *buf, size_t count);

     int recv_s(int sockfd, void *buf, size_t len, int flags);
 
     int  recvfrom_s(int  sockfd,  void  *buf,  size_t len, int flags,
                     struct sockaddr *from, socklen_t *fromlen);
 
     int recvmsg_s(int sockfd, struct msghdr *msg, int flags);                       


     /* set socket parameters (joining multicast group, ..)*/
     int setsockopt_s(int sockfd, int level, int optname,  const  void
                      *optval, socklen_t optlen);

     /* The socket call corresponding to the TCP mode of operation (client &
        server) have not been defined as of now */

     /* closing a socket*/
     int close_s(int sockfd);

     /* set the policy file to be used by the group object corresponding to 
        the socket "sockfd" */
     int set_policy(int sockfd, const char *filename);

#ifdef SHARAD_SOURCE
}
#endif
#endif












